"""The m3 server, which handles all requests over WebSockets and delegates the actual work to the
pseucparser and to the Z3Compiler"""

import ssl
import itertools
import functools
import time
import json
import asyncio
import multiprocessing
import logging
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pyparsing import ParseException

from autobahn.asyncio.websocket import WebSocketServerProtocol, \
    WebSocketServerFactory

from pseuco_parser import createParser, UnsupportedPseuCoException
from z3compiler import Z3Compiler
from communication import (Explanation, MemoryModel, InstructionId, OrderAssertion,
                           WriteSeenAssertion, ReadValueAssertion, WriteValueAssertion)


def get_cfg_for_program(program):
    """Creates a cfg representation of the given program"""
    ret = {}
    counter = 0
    for agent_name in itertools.chain.from_iterable(
            program.startedAgentNames.asList()):
        agent = program.initialization.agent_declarations[agent_name]
        body = agent.functionBody
        cfg_predecessors = []
        cfc_graph_representation = []
        ret["init0"] = program.initialization.create_node_data()
        for statement in body.statements:
            cfc_graph_representation += statement.create_node_data(
                cfg_predecessors)
            cfg_predecessors = statement.get_cfg_exit_nodes()
        ret[agent_name + str(counter+1)] = cfc_graph_representation
        counter += 1
    return ret


class BookKeeper:
    """Keeps track of how many ressources a client is using,
       Currently broken, and thus not used
    """
    def __init__(self, score_limit=1000):
        """Sets up the bookkeeping"""
        self.ip2score = defaultdict(int)
        self.ip2last_access = {}
        self.time_of_last_reset = time.monotonic()
        self.limit = score_limit

    def _improve_score(self):
        time_since_reset = time.monotonic() - self.time_of_last_reset
        score_reduction = int(time_since_reset) / (60000 * 30)  # 15 minutes
        if time_since_reset <= 0:
            return
        self.time_of_last_reset = time.monotonic()
        improvement = int(time_since_reset)
        removable_users = []
        for ip, score in self.ip2score.items():
            new_score = max(0, score - 100*score_reduction)
            if new_score == 0:
                removable_users.append(ip)
            else:
                self.ip2score[ip] = new_score
        for user in removable_users:
            del self.ip2last_access[user]
            del self.ip2score[user]

    def caused_timeout(self, ip):
        """Called when a user caused a timeout of the z3 solver"""

    def track_and_check_visit(self, ip):
        """Call when user visits site"""
        self._improve_score()
        self.ip2last_access[ip] = time.monotonic()
        return self.ip2score[ip] > self.limit

    def update_score(self, ip):
        last_access = time.monotonic()
        time_spent = int(self.ip2last_access[ip] - last_access)
        self.ip2score[ip] += (time_spent+1)*50

book_keeper = BookKeeper()


class M3ServerProtocoll(WebSocketServerProtocol):

    def _ack_nack_helper(func):
        """Easier message handling"""
        @functools.wraps(func)
        async def wrapper(self, *args, **kwargs):
            success, id, explanation = await func(self, *args, **kwargs)
            logging.info(f"{func.__name__}, {success}, {id}, {explanation}")
            if success:
                return {
                    "type": "ack",
                    "id": id
                }
            else:
                return {
                    "type": "nack",
                    "id": id,
                    "explanation": explanation
                }

        return wrapper


    executor = ThreadPoolExecutor(max_workers=multiprocessing.cpu_count(),
                                  thread_name_prefix="z3_task")

    mem_model = {
        "seqConsist": MemoryModel.SequentialConsistency,
        "weak": MemoryModel.WeakConsistency
    }

    def __init__(self):
        super().__init__()
        self.handle = {
            "code": self._onCodeMessage,
            "getAllExecutions": self._onGetAllExecutions,
            "WriteSeenAdded": self._onWriteSeenAdded,
            "setMemoryModel": self._onsetMemoryModel,
            "valueSet": self._onValueSet,
            "writeValueSet": self._onWriteValueSet,
            "OrderAdded": self._onOrderAdded,
            "undoAssertion": self._onUndoAssertion,
        }
        self.outstanding_work = None
        self.compiler = None

    async def _onUndoAssertion(self, decoded_message):
        id_ = decoded_message["id"]
        self.compiler.remove_assertion(id_)
        reply = {
            "type": "removedAssertion",
            "id": id_
        }
        return reply

    async def _onCodeMessage(self, decoded_message):
        try:
            tinyPseucoParser = createParser()
            program = tinyPseucoParser.parseString(decoded_message["code"])
            reply = {
                "cfg": get_cfg_for_program(program),
                "type": "parseResult"
            }
            loop = asyncio.get_event_loop()
            self.outstanding_work = loop.run_in_executor(
                M3ServerProtocoll.executor,
                Z3Compiler,
                program,
            )
        except (UnsupportedPseuCoException, ParseException) as e:
            reply = {
                "type": "error",
                "errorMessage": str(e),
            }
        return reply

    async def _onGetAllExecutions(self, decoded_message):
        # loop = asyncio.get_event_loop()
        # allReads = await loop.run_in_executor(
        #     M3ServerProtocoll.executor,
        #     self.compiler.list_executions,
        # )
        for execution in self.compiler.list_executions():
            yield {
                "type": "newRead",
                "read": execution,
            }
        yield {
            "type": "allReadsFinished",
        }

    @_ack_nack_helper
    async def _onsetMemoryModel(self, decoded_message):
        mem_model = MemoryModel(decoded_message["memoryModel"])
        return (*self.compiler.set_memory_model(mem_model), None)

    @_ack_nack_helper
    async def _onWriteSeenAdded(self, decoded_message):
        read = InstructionId(decoded_message["seen"]["agent"],
                             decoded_message["seen"]["statement"])
        write = InstructionId(decoded_message["write"]["agent"],
                              decoded_message["write"]["statement"])
        write_seen = WriteSeenAssertion(writer=write,
                                        id=decoded_message["id"],
                                        reader=read,
                                        variable=decoded_message["varname"])
        sat, explanation = self.compiler.add_constraint(write_seen)
        return sat, decoded_message["id"], explanation

    @_ack_nack_helper
    async def _onOrderAdded(self, decoded_message):
        prev = InstructionId(decoded_message["prev"]["agent"],
                             decoded_message["prev"]["statement"])
        succ = InstructionId(decoded_message["next"]["agent"],
                             decoded_message["next"]["statement"])
        order = OrderAssertion(successor=succ, predecessor=prev, id=decoded_message["id"])
        sat, explanation = self.compiler.add_constraint(order)
        return sat, decoded_message["id"], explanation

    @_ack_nack_helper
    async def _onValueSet(self, decoded_message):
        value = int(decoded_message["value"])
        variable = decoded_message["varname"]
        instruction = InstructionId(decoded_message["statement"]["agent"],
                                    decoded_message["statement"]["statement"]
                                   )
        read_value = ReadValueAssertion(instruction=instruction,
                                        variable=variable,
                                        value=value,
                                        id=decoded_message["id"])
        sat, explanation = self.compiler.add_constraint(read_value)
        return sat, decoded_message["id"], explanation

    @_ack_nack_helper
    async def _onWriteValueSet(self, decoded_message):
        value = int(decoded_message["value"])
        variable = decoded_message["varname"]
        instruction = InstructionId(decoded_message["statement"]["agent"],
                                    decoded_message["statement"]["statement"]
                                   )
        write_value = WriteValueAssertion(instruction=instruction,
                                          variable=variable,
                                          value=value,
                                          id=decoded_message["id"])
        sat, explanation = self.compiler.add_constraint(write_value)
        return sat, decoded_message["id"], explanation


    async def onMessage(self, payload, isBinary):
        assert not isBinary, "Where did the binary message come from" #  turn off in production
        decoded_message = json.loads(payload.decode('utf8'))
        if self.outstanding_work:
            self.compiler =  await self.outstanding_work
            self.outstanding_work = None
        try:
            message_type = decoded_message["type"]
            reply = await self.handle[message_type](decoded_message)
        except KeyError:
            reply = {
                "type": "error",
                "errorMessage": "%s is not a supported message type" % decoded_message["type"]
            }
        self.sendMessage(json.dumps(reply).encode('utf8'))


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="The m3 server")
    parser.add_argument("--port", "-p", type=int, help="The port, defaults to 5678", default=5678)
    parser.add_argument("--hostname", "-n", type=str, help="The hostname, defaults to 127.0.0.1",
                        default="127.0.0.1")
    parser.add_argument('-v', '--verbose', action='count', default=0)
    args = parser.parse_args()

    host = args.hostname
    port = args.port
    if args.verbose > 0:
        logging.basicConfig(level=logging.INFO)

    factory = WebSocketServerFactory("wss://%s:%d" % (host, port))
    factory.protocol = M3ServerProtocoll

    sc = None
    try:
        sc = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        sc.load_cert_chain('/home/m3-deploy/certificates/cert.pem', '/home/m3-deploy/certificates/privkey.pem')
    except Exception as e:
        print("couldn't open ssl certs")
        print(str(e))
        sc = None

    loop = asyncio.get_event_loop()
    coro = loop.create_server(factory, host, port, ssl=sc)

    server = loop.run_until_complete(coro)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.close()
        loop.close()
